// Sweeps through all voxel values and renormalizes them to the Hounsfield scale.
__kernel void convert_to_hu_cl(
	__global float *img,
	int npix,
	int offset,
	int nvoxels
) {
	int p = get_global_id(0);

	if ( p < nvoxels)
		img[p] = (149.70059880239520958083832335329 * img[p]) - 1000;
}

// Define image/texture sampling parameters
const sampler_t dev_img_sampler = CLK_NORMALIZED_COORDS_FALSE | CLK_ADDRESS_NONE | CLK_FILTER_NEAREST;

// FDK kernel 
__kernel void kernel_fdk(
	__global float *dev_vol,
	__read_only image2d_t dev_img,
	__constant float *dev_matrix,
	__constant float4 *nrm,
	__constant float4 *vol_offset,
	__constant float4 *vol_pix_spacing,
	__constant int4 *vol_dim,
	__constant float2 *ic,
	__constant int2 *img_dim,
	__constant float *sad,
	__constant float *scale,
	__constant int4 *offset,
	__global int4 *ndevice
) {
	uint i = get_global_id(0);
	uint j = get_global_id(1);
	uint k = get_global_id(2);

	if (i >= (*ndevice).x || j >= (*ndevice).y || k >= (*ndevice).z)
		return;

	// Index row major into the volume
	long vol_idx = i + (j * (*vol_dim).x) + (k * (*vol_dim).x * (*vol_dim).y);
	vol_idx -= (*offset).w;

	i += (*offset).x;
	j += (*offset).y;
	k += (*offset).z;

	// Get volume value from global memory
	float dev_vol_value = dev_vol[vol_idx];

	// offset volume coords
	float4 vp;
	vp.x = (*vol_offset).x + (i * (*vol_pix_spacing).x);	// Compiler should combine into 1 FMAD.
	vp.y = (*vol_offset).y + (j * (*vol_pix_spacing).y);	// Compiler should combine into 1 FMAD.
	vp.z = (*vol_offset).z + (k * (*vol_pix_spacing).z);	// Compiler should combine into 1 FMAD.

	// matrix multiply
	float4 ip;
	ip.x = (dev_matrix[0] * vp.x) + (dev_matrix[1] * vp.y) + (dev_matrix[2] * vp.z) + dev_matrix[3];
	ip.y = (dev_matrix[4] * vp.x) + (dev_matrix[5] * vp.y) + (dev_matrix[6] * vp.z) + dev_matrix[7];
	ip.z = (dev_matrix[8] * vp.x) + (dev_matrix[9] * vp.y) + (dev_matrix[10] * vp.z) + dev_matrix[11];

	// Change coordinate systems
	ip.x = (*ic).x + ip.x / ip.z;
	ip.y = (*ic).y + ip.y / ip.z;

	// Get pixel location from 2D image
	int2 pos;
	pos.y = convert_int_rtn(ip.x);
	pos.x = convert_int_rtn(ip.y);

	// Clip against image dimensions
	if (pos.x < 0 || pos.x >= (*img_dim).x || pos.y < 0 || pos.y >= (*img_dim).y)
		return;

	// Get pixel from texture memory
	float4 voxel_data = read_imagef(dev_img, dev_img_sampler, pos);
	//float4 voxel_data = (0.0, 0.0, 0.0, 0.0);

	// Dot product
	float s = ((*nrm).x * vp.x) + ((*nrm).y * vp.y) + ((*nrm).z * vp.z);

	// Conebeam weighting factor
	s = (*sad) - s;
	s = ((*sad) * (*sad)) / (s * s);

	// Place it into the volume
	dev_vol[vol_idx] = dev_vol_value + ((*scale) * s * voxel_data.x);
}
