#ifndef _FDK_OPENCL_H_
#define _FDK_OPENCL_H_

#include <stdio.h>
#include <stdlib.h>
#include <cstdlib>
#include <math.h>
#include <string.h>
#include <fstream>
#include <oclUtils.h>

// D A T A   S T R U C T U R E S //////////////////////////////////////////
struct int2 {
	int x;
	int y;
};

struct int4 {
	int x;
	int y;
	int z;
	int w;
};

struct float2 {
	float x;
	float y;
};

struct float4 {
	float x;
	float y;
	float z;
	float w;
};

enum Pixel_Type
{
    PT_UNDEFINED,
    PT_UCHAR,
    PT_SHORT,
    PT_UINT32,
    PT_FLOAT,
    PT_VF_FLOAT_INTERLEAVED,
    PT_VF_FLOAT_PLANAR
};

typedef struct MGHCBCT_Options_struct MGHCBCT_Options;
struct MGHCBCT_Options_struct {
    int first_img;
    int skip_img;
    int last_img;
    int resolution[3];
    float vol_size[3];
    float scale;
    const char *input_dir;
    const char *output_file;
};

typedef struct cb_image CB_Image;
struct cb_image
{
    int dim[2];         /* dim[0] = cols
			   dim[1] = rows */
    double ic[2];	/* Image Center
			   ic[0] = x
			   ic[1] = y     */
    double matrix[12];	// Projection matrix
    double sad;		// Distance: Source To Axis
    double sid;		// Distance: Source to Image
    double nrm[3];	// Ray from image center to source
    float *img;		// Pixel data
};

typedef struct volume Volume;
struct volume
{
    int dim[3];			// x, y, z Dims
    int npix;			// # of pixels in volume
				// = dim[0] * dim[1] * dim[2] 
    float offset[3];
    float pix_spacing[3];	// voxel spacing
    float direction_cosines[9];

    enum Pixel_Type pix_type;	// Voxel Data type
    int pix_size;
    void *img;			// Voxel Data

    float xmin;	// Minimum X Value in Volume
    float xmax; // Maximum X Value in Volume
    float ymin; // Minimum Y Value in Volume
    float ymax; // Maximum Y Value in Volume
    float zmin; // Minimum Z Value in Volume
    float zmax; // Maximum Z Value in Volume
};

struct kernel_args_fdk
{
	int2 img_dim;
	float2 ic;
	float4 nrm;
	float sad;
	float sid;
	float scale;
	float4 vol_offset;
	int4 vol_dim;
	float4 vol_pix_spacing;
	float matrix[12];
	int4 v_offset;
	int4 v_device;
};

// P R O T O T Y P E S ////////////////////////////////////////////////////
cl_ulong executionTime(cl_event &event);
void free_cb_image (CB_Image *cbi);
void fwrite_block (void *buf, size_t size, size_t count, FILE *fp);
void parse_args (MGHCBCT_Options *options, int argc, char *argv[]);
void print_usage(void);
void set_default_options (MGHCBCT_Options *options);
void write_mha (char *filename, Volume *vol);
CB_Image* get_image (MGHCBCT_Options *options, int image_num);
CB_Image* load_cb_image (char *img_filename, char *mat_filename);
Volume* my_create_volume (MGHCBCT_Options *options);
Volume* volume_create (int *dim, float *offset, float *pix_spacing, enum Pixel_Type pix_type, float *direction_cosines, int min_size);
cl_uint get_device_warp(cl_device_id device);
int gcf(cl_int a, cl_int b);
int main(int argc, char *argv[]);
float reconstruct_conebeam_and_convert_to_hu (Volume *vol, MGHCBCT_Options *options);

#endif

